# Copyright (c) 2020  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Trivial tests for the ValidationError class."""

import trivval


def test_error_string() -> None:
    """Test the stringification of the error."""
    messages = (
        ["word"],
        ["two words"],
        ["two", "words"],
        ["slash/phrase", "and: more"],
    )

    for msg in messages:
        joined = "\t".join(msg)
        err = trivval.ValidationError([], joined)
        assert str(err) == joined

    for key in messages:
        for msg in messages:
            slashed = "/".join(key)
            joined = "\t".join(msg)
            err = trivval.ValidationError(key, joined)
            assert str(err) == f"{joined}: {slashed}"



# Copyright (c) 2020  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Data samples for the validation library tests."""


SCHEMA_1_0 = {
    "node": {
        "hostname": str,
        "flavor": set(("desktop", "laptop", "server")),
        "services": {
            "launcher": {"timestamp": int},
            "server": {"timestamp": int, "instances": int},
        },
    },
    "users": {"*": {"first_name": str, "last_name": str, "id": int}},
    "tags": [str],
}

SCHEMA_1_1 = {
    "node": {
        "hostname": str,
        "flavor": set(("desktop", "laptop", "server")),
        "services": {
            "launcher": {"timestamp": int},
            "server": {"timestamp": int, "instances": int},
        },
    },
    "users": {"*": {"first_name": str, "last_name": str, "id": int, "?nickname": str}},
    "tags": [str],
}

SCHEMA_2_0 = {
    "nodes": {
        "*": {
            "hostname": str,
            "flavor": set(("laptop", "server")),
            "services": {
                "launcher": {"timestamp": int},
                "server": {"timestamp": int, "instances": int},
            },
        },
    },
    "users": {
        "*": {
            "first_name": str,
            "last_name": str,
            "id": int,
            "?nickname": str,
        }
    },
    "tags": [str],
}

SCHEMA = {
    (1, 0): SCHEMA_1_0,
    (1, 1): SCHEMA_1_1,
    (2, 0): SCHEMA_2_0,
}

VALUES_1_0 = [
    {
        "node": {
            "hostname": "straylight",
            "flavor": "laptop",
            "services": {
                "launcher": {"timestamp": 1000},
                "server": {"timestamp": 1001, "instances": 4},
            },
        },
        "users": {
            "jrl": {
                "first_name": "J. Random",
                "last_name": "Lucretius",
                "id": 42,
            },
            "centre": {"first_name": "Aza", "last_name": "Thoth", "id": 616},
        },
        "tags": ["foo", "bar"],
    },
    {
        "node": {
            "hostname": "feylight",
            "flavor": "desktop",
            "services": {
                "launcher": {"timestamp": 3000},
                "server": {"timestamp": 3005, "instances": 4},
            },
        },
        "users": {
            "guest": {
                "first_name": "Mister",
                "last_name": "Apollinax",
                "id": 201,
            },
        },
        "tags": [],
    },
]

VALUES_1_1 = [
    {
        "node": {
            "hostname": "straylight",
            "flavor": "server",
            "services": {
                "launcher": {"timestamp": 1000},
                "server": {"timestamp": 1001, "instances": 4},
            },
        },
        "users": {
            "jrl": {
                "first_name": "J. Random",
                "last_name": "Lucretius",
                "id": 42,
                "nickname": "jay",
            },
            "centre": {"first_name": "Aza", "last_name": "Thoth", "id": 616},
        },
        "tags": ["foo", "bar"],
    },
    {
        "node": {
            "hostname": "feylight",
            "flavor": "desktop",
            "services": {
                "launcher": {"timestamp": 3000},
                "server": {"timestamp": 3005, "instances": 4},
            },
        },
        "users": {
            "guest": {
                "first_name": "Mister",
                "last_name": "Apollinax",
                "id": 201,
                "nickname": "paul",
            },
        },
        "tags": [],
    },
]

VALUES_2_0 = [
    {
        "nodes": {
            "laptop": {
                "hostname": "straylight",
                "flavor": "laptop",
                "services": {
                    "launcher": {"timestamp": 1000},
                    "server": {"timestamp": 1001, "instances": 4},
                },
            },
        },
        "users": {
            "jrl": {
                "first_name": "J. Random",
                "last_name": "Lucretius",
                "id": 42,
                "nickname": "jay",
            },
            "centre": {"first_name": "Aza", "last_name": "Thoth", "id": 616},
        },
        "tags": ["foo", "bar"],
    },
    {
        "nodes": {
            "local": {
                "hostname": "feylight",
                "flavor": "server",
                "services": {
                    "launcher": {"timestamp": 3000},
                    "server": {"timestamp": 3005, "instances": 4},
                },
            },
        },
        "users": {
            "guest": {
                "first_name": "Mister",
                "last_name": "Apollinax",
                "id": 201,
                "nickname": "paul",
            },
        },
        "tags": [],
    },
]

VALUES = {
    (1, 0): VALUES_1_0,
    (1, 1): VALUES_1_1,
    (2, 0): VALUES_2_0,
}



# Copyright (c) 2020 - 2022  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Validate the format of the Markdown documentation files."""

from __future__ import annotations

import dataclasses
import pathlib
import re

from typing import Any, Iterator, Union

import mistune  # type: ignore

from packaging import version as pversion

import trivval


@dataclasses.dataclass(frozen=True)
class MDHeadingParams:
    """A Markdown heading's parameters."""

    level: int


@dataclasses.dataclass(frozen=True)
class MDHeading:
    """A Markdown heading."""

    type: str
    text: str
    params: MDHeadingParams


@dataclasses.dataclass(frozen=True)
class MDListItemParams:
    """A Markdown list item's parameters."""

    depth: int


@dataclasses.dataclass(frozen=True)
class MDListItem:
    """A Markdown list item."""

    type: str
    children: list["MDObject"]
    params: MDListItemParams


@dataclasses.dataclass(frozen=True)
class MDListParams:
    """A Markdown list's parameters."""

    ordered: bool
    depth: int
    start: Any


@dataclasses.dataclass(frozen=True)
class MDList:
    """A Markdown list."""

    type: str
    children: list[MDListItem]
    params: MDListParams


@dataclasses.dataclass(frozen=True)
class MDBlockText:
    """A Markdown text string."""

    type: str
    text: str


MDObject = Union[MDBlockText, MDHeading, MDList]

RE_HEADING = re.compile(
    r""" ^
        (?P<version> \d+ \. \d+ \. \d+ ) \s+
        [(]
        (?:
            (?P<date> \d{4} - \d{2} - \d{2} )
            |
            (?P<not_yet> not [ ] yet )
        )
        [)]
    $ """,
    re.X,
)


def parse_object(obj: dict[str, Any]) -> MDObject:
    """Parse an object returned by mistune into a dataclass."""

    def parse_block_text(obj: dict[str, Any]) -> MDObject:
        """Parse a block of text."""
        assert obj["type"] == "block_text"
        return MDBlockText(type=str(obj["type"]), text=str(obj["text"]))

    def parse_heading(obj: dict[str, Any]) -> MDObject:
        """Parse a heading."""
        assert obj["type"] == "heading"
        assert len(obj["params"]) == 1
        return MDHeading(
            type=str(obj["type"]),
            text=str(obj["text"]),
            params=MDHeadingParams(level=int(obj["params"][0])),
        )

    def parse_list_item(obj: dict[str, Any]) -> MDListItem:
        """Parse a list item."""
        assert obj["type"] == "list_item"
        assert len(obj["params"]) == 1
        return MDListItem(
            type=str(obj["type"]),
            children=[parse_object(item) for item in obj["children"]],
            params=MDListItemParams(depth=int(obj["params"][0])),
        )

    def parse_list(obj: dict[str, Any]) -> MDObject:
        """Parse a list."""
        assert obj["type"] == "list"
        assert len(obj["params"]) == 3
        return MDList(
            type=str(obj["type"]),
            children=[parse_list_item(item) for item in obj["children"]],
            params=MDListParams(
                ordered=bool(obj["params"][0]),
                depth=int(obj["params"][1]),
                start=obj["params"][2],
            ),
        )

    handlers = {
        "block_text": parse_block_text,
        "heading": parse_heading,
        "list": parse_list,
    }

    return handlers[obj["type"]](obj)


def test_changelog() -> None:
    """Test the format of the CHANGES.md file."""
    text = pathlib.Path("CHANGES.md").read_text(encoding="UTF-8")
    parser = mistune.BlockParser()
    state: dict[str, Any] = {}
    raw: list[dict[str, Any]] = parser.parse(text, state)
    # pylint: disable-next=not-an-iterable
    res = [parse_object(item) for item in raw]

    def validate_list(items: MDList) -> None:
        """Recursively validate a list."""
        for listitem in items.children:
            assert isinstance(listitem, MDListItem)
            for item in listitem.children:
                if isinstance(item, MDBlockText):
                    continue

                assert isinstance(item, MDList)
                validate_list(item)

    def validate_entries(lst: Iterator[MDObject]) -> None:
        """Validate the changelog entries."""
        last_version: str | None = None
        while True:
            try:
                heading = next(lst)
            except StopIteration:
                assert last_version is not None
                return
            assert isinstance(heading, MDHeading) and heading.params.level == 2

            data = RE_HEADING.match(heading.text)
            assert data, heading.text
            version = data.group("version")
            if last_version is None:
                assert version == trivval.VERSION
            else:
                assert pversion.parse(version) < pversion.parse(last_version)
                assert data.group("date")
            last_version = version

            items = next(lst)
            assert isinstance(items, MDList)
            validate_list(items)

    assert isinstance(res[0], MDHeading) and res[0].params.level == 1
    validate_entries(iter(res[1:]))


def test_readme() -> None:
    """Just test that README.md may be parsed."""
    text = pathlib.Path("README.md").read_text(encoding="UTF-8")
    parser = mistune.BlockParser()
    parser.parse(text, {})



# Copyright (c) 2020 - 2022  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Simple tests for the validation routine."""

from __future__ import annotations

import json

from unittest import mock

from typing import Any

import pytest

import trivval

from . import data as tdata


def do_test(value: dict[str, Any], schema: dict[str, Any], backwards: bool = False) -> None:
    """Run the tests for a specific value against a schema."""

    def test_extra_succeed() -> None:
        trivval.validate_dict(value, schema, 0)

    def test_extra_fail() -> None:
        with pytest.raises(trivval.ValidationError) as err:
            trivval.validate_dict(value, schema, 0)
        assert err.value.path[-1] != "a,more"
        assert err.value.err.startswith("extra")

    test_extra = test_extra_fail if backwards else test_extra_succeed

    trivval.validate_dict(value, schema, trivval.FLAG_ALLOW_EXTRA)
    test_extra()

    if "node" in value:
        add_to = value["node"]
        add_path = ["node"]
    else:
        add_key = next(iter(value["nodes"].keys()))
        add_to = value["nodes"][add_key]
        add_path = ["nodes", add_key]
    assert "more" not in add_to and "a" not in add_to
    add_to.update({"a": "lot", "more": "stuff"})
    trivval.validate_dict(value, schema, trivval.FLAG_ALLOW_EXTRA)
    with pytest.raises(trivval.ValidationError) as err:
        trivval.validate_dict(value, schema, 0)
    if not backwards:
        assert err.value.path == add_path + ["a,more"]
        assert err.value.err.startswith("extra")

    del add_to["a"]
    del add_to["more"]
    trivval.validate_dict(value, schema, trivval.FLAG_ALLOW_EXTRA)
    test_extra()

    missing = add_to.pop("hostname")

    with pytest.raises(trivval.ValidationError) as err:
        trivval.validate_dict(value, schema, trivval.FLAG_ALLOW_EXTRA)
    assert err.value.path == add_path + ["hostname"]
    assert err.value.err.startswith("missing")

    with pytest.raises(trivval.ValidationError) as err:
        trivval.validate_dict(value, schema, 0)
    if not backwards:
        assert err.value.path == add_path + ["hostname"]
        assert err.value.err.startswith("missing")

    add_to["hostname"] = missing
    trivval.validate_dict(value, schema, trivval.FLAG_ALLOW_EXTRA)
    test_extra()


def test_exact() -> None:
    """Test against the same version."""
    for version in sorted(tdata.VALUES.keys()):
        schema = tdata.SCHEMA[version]
        for value in tdata.VALUES[version]:
            do_test(value, schema)


def test_backwards() -> None:
    """Test against an earlier version."""
    for current_version in sorted(tdata.VALUES.keys()):
        for earlier_version in sorted(
            ver
            for ver in tdata.SCHEMA
            if ver[0] == current_version[0] and ver[1] < current_version[1]
        ):
            schema = tdata.SCHEMA[earlier_version]
            for value in tdata.VALUES[current_version]:
                do_test(value, schema, backwards=True)


def test_breaking() -> None:
    """Make sure validation fails after a breaking change."""
    for current_version in sorted(tdata.VALUES.keys()):
        for other_version in sorted(ver for ver in tdata.SCHEMA if ver[0] != current_version[0]):
            schema = tdata.SCHEMA[other_version]
            for value in tdata.VALUES[current_version]:
                with pytest.raises(trivval.ValidationError):
                    trivval.validate_dict(value, schema, 0)


def test_schema() -> None:
    """Make sure validation succeeds with all schemas."""
    called: list[tuple[dict[str, Any], trivval.SchemaType, int] | None] = [None]

    def mock_validate_dict(value: dict[str, Any], schema: trivval.SchemaType, flags: int) -> None:
        called.append((value, schema, flags))

    for version, values in tdata.VALUES.items():
        v_format = {"version": {"major": version[0], "minor": version[1]}}
        for value in values:
            augmented = dict(value)
            assert "format" not in augmented
            augmented["format"] = v_format

            trivval.validate(augmented, tdata.SCHEMA, 0)

            assert augmented["format"] is v_format
            del augmented["format"]
            assert augmented == value

            augmented["format"] = v_format
            called = [None]

            with mock.patch("trivval.validate_dict", new=mock_validate_dict):
                trivval.validate(augmented, tdata.SCHEMA, 0)

            assert called[1] is not None
            assert called[1][0] == augmented
            called.pop(1)
            assert called == [
                None,
                (value, tdata.SCHEMA[version], 0),
            ]


def test_exact_json() -> None:
    """Test against the same version after a round trip through JSON."""
    for version in sorted(tdata.VALUES.keys()):
        schema = tdata.SCHEMA[version]
        for value in tdata.VALUES[version]:
            do_test(json.loads(json.dumps(value)), schema)


def test_optional() -> None:
    """Test "?key" not present or present as a null value."""
    schema = {"first": str, "?second": int}

    trivval.validate_dict({"first": "hello"}, schema, 0)
    trivval.validate_dict({"first": "hello", "second": None}, schema, 0)
    trivval.validate_dict({"first": "goodbye", "second": 42}, schema, 0)

    with pytest.raises(trivval.ValidationError):
        trivval.validate_dict({"second": 42}, schema, 0)

    with pytest.raises(trivval.ValidationError):
        trivval.validate_dict({"first": "1", "second": "2"}, schema, 0)
